/*
 * Copyright (c) 2019 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-BSD-5-Clause-Nordic
 */

#include <zephyr.h>
#include <debug/ppi_trace.h>
#include <counter.h>
#include <nrf_rtc.h>
#include <nrf_clock.h>
#include <device.h>
#include <logging/log.h>
#include <nrfx_dppi.h>
#include <nrf_dppi.h>
#include <nrf_timer.h>
#include <nrfx_timer.h>
#include <nrf_gpiote.h>
#include <nrfx_gpiote.h>
#include <nrf_gpio.h>
#include <nrfx.h>

LOG_MODULE_REGISTER(app);

#define GPIO_INPUT_PIN_NUMBER 13

u32_t C = 0;
u32_t pre_value = 0;
u32_t cur_value = 0;
u32_t f_value = 0;

#define TIMER_TIMER NRF_TIMER_2
#define COUNTER_TIMER NRF_TIMER_1

static nrfx_timer_t timer_1 = NRFX_TIMER_INSTANCE(1);
static nrfx_timer_t timer_2 = NRFX_TIMER_INSTANCE(2);

void timer_1_dummy_handler(nrf_timer_event_t evt_type, void * p_context){}

void timer_2_event_handler(nrf_timer_event_t evt_type, void * p_context)
{
    if (evt_type == NRF_TIMER_EVENT_COMPARE0) {
        cur_value = NRF_TIMER1_NS->CC[0];
        f_value = cur_value - pre_value;
        pre_value = cur_value;
        printk("%d, %d, %d\n", pre_value, cur_value, f_value);
    }
}

void gpiote_dummy_handler(nrfx_gpiote_pin_t pin, nrf_gpiote_polarity_t action){}

static void gpiote_setup()
{
    u32_t counter_task_addr;
    u32_t gpiote_evt_addr;
    nrf_dppi_channel_group_t dppi_channel_counter;
    int err_code;
    nrfx_gpiote_in_config_t config = NRFX_GPIOTE_CONFIG_IN_SENSE_LOTOHI(true);

    err_code = nrfx_gpiote_in_init(GPIO_INPUT_PIN_NUMBER, &config, gpiote_dummy_handler);
    __ASSERT(err_code == NRFX_SUCCESS, "");

    err_code = nrfx_dppi_channel_alloc(&dppi_channel_counter);
    __ASSERT(err_code == NRFX_SUCCESS, "");

    counter_task_addr = nrfx_timer_task_address_get(&timer_1, NRF_TIMER_TASK_COUNT);
    gpiote_evt_addr = nrfx_gpiote_in_event_addr_get(GPIO_INPUT_PIN_NUMBER);
    printk("gpiote_evt_addr: 0x%08x\n", gpiote_evt_addr);
    printk("NRF_GPIOTE_EVENTS_IN_0: 0x%08x\n", NRF_GPIOTE_EVENTS_IN_0);
    printk("NRF_GPIOTE1_NS: 0x%08x\n", NRF_GPIOTE1_NS);
    printk("gpiote_evt_addr - NRF_GPIOTE0_S: 0x%08x\n", gpiote_evt_addr - (u32_t)NRF_GPIOTE1_NS);

    //err_code = nrf_drv_ppi_channel_assign(ppi_channel_counter, gpiote_evt_addr, counter_task_addr);
    // gpiote_evt_addr - (u32_t)NRF_GPIOTE0_S = NRF_GPIOTE_EVENTS_IN_0, is this always true?
    // There probably is a better way of converting a pin in event to a gpiote channel in event, but I do not know it
    nrf_gpiote_publish_set(gpiote_evt_addr - (u32_t)NRF_GPIOTE1_NS, dppi_channel_counter);
    nrf_timer_subscribe_set(NRF_TIMER1_NS, NRF_TIMER_TASK_COUNT, dppi_channel_counter);
    __ASSERT(err_code == NRFX_SUCCESS, "");

    err_code = nrfx_dppi_channel_enable(dppi_channel_counter);
    __ASSERT(err_code == NRFX_SUCCESS, "");

    nrfx_gpiote_in_event_enable(GPIO_INPUT_PIN_NUMBER, true); // Enable EVENT without the interrupt

    u32_t timer_evt_addr;
    nrf_dppi_channel_group_t dppi_channel_timer;

    err_code = nrfx_dppi_channel_alloc(&dppi_channel_timer);
    __ASSERT(err_code == NRFX_SUCCESS, "");

    counter_task_addr = nrfx_timer_task_address_get(&timer_1, NRF_TIMER_TASK_CAPTURE0);
    timer_evt_addr = nrfx_timer_event_address_get(&timer_2, NRF_TIMER_EVENT_COMPARE0);

    //err_code = nrfx_dppi_channel_assign(dppi_channel_timer, timer_evt_addr, counter_task_addr);
    nrf_timer_publish_set(NRF_TIMER2_NS, NRF_TIMER_EVENT_COMPARE0, dppi_channel_timer);
    nrf_timer_subscribe_set(NRF_TIMER1_NS, NRF_TIMER_TASK_CAPTURE0, dppi_channel_timer);
    __ASSERT(err_code == NRFX_SUCCESS, "");

    err_code = nrfx_dppi_channel_enable(dppi_channel_timer);
    __ASSERT(err_code == NRFX_SUCCESS, "");
}

void main()
{
    #if 0
    // Start HFCLK with external crystal
    NRF_CLOCK->TASKS_HFCLKSTART = 1;
    while (NRF_CLOCK->EVENTS_HFCLKSTARTED == 0);
    NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;
    #endif

    u32_t time_ms = 1000; // 1sek
    u32_t time_ticks_ms;
    int err_code;

    u32_t value;

    //err_code = nrfx_dppi_init();
    //ASSERT(err_code == NRFX_SUCCESS);

    err_code = nrfx_gpiote_init();
    __ASSERT(err_code == NRFX_SUCCESS, "");

    nrfx_timer_config_t timer_1_cfg = NRFX_TIMER_DEFAULT_CONFIG;
    timer_1_cfg.mode = TIMER_MODE_MODE_Counter;
    timer_1_cfg.bit_width = NRF_TIMER_BIT_WIDTH_32;
    nrfx_timer_uninit(&timer_1); // Zephyr has already initialized the timer, so we need to uninitialize it before we can configure it
    err_code = nrfx_timer_init(&timer_1, &timer_1_cfg, timer_1_dummy_handler);
    if (err_code != NRFX_SUCCESS) {
        printk("error %x @ %d\n", err_code, __LINE__);
        return;
    }

    nrfx_timer_config_t timer_2_cfg = NRFX_TIMER_DEFAULT_CONFIG;
    timer_2_cfg.bit_width = NRF_TIMER_BIT_WIDTH_32;
    nrfx_timer_uninit(&timer_2);
    err_code = nrfx_timer_init(&timer_2, &timer_2_cfg, timer_2_event_handler);
    __ASSERT(err_code == NRFX_SUCCESS, "");

    time_ticks_ms = nrfx_timer_ms_to_ticks(&timer_2, time_ms);
    nrfx_timer_extended_compare(&timer_2, NRF_TIMER_CC_CHANNEL0, time_ticks_ms, NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK, true);

    gpiote_setup();

    nrfx_timer_enable(&timer_1);
    
    nrfx_timer_enable(&timer_2);

    printk("Everything is set up\n");
    //while (1) {}
}